import functools
import json
from datetime import datetime

import click
import matplotlib.pyplot as plt
import requests
from matplotlib import animation


class GraphRealTime:
    def __init__(self, query: str, elasticsearch_url: str, index: str, field: str, refresh_interval: int,
                 xs_param: list = None, ys_param: list = None):
        if query.startswith("@"):
            with open(query[1:]) as f:
                self.query = json.loads(f.read())
        else:
            self.query = json.loads(query)
        self.field = field
        self.elasticsearch_endpoint = '%s/%s/_search?size=0' % (elasticsearch_url, index)
        self.refresh_interval = refresh_interval
        self.xs = [] if xs_param is None else xs_param
        self.ys = [] if xs_param is None else ys_param

        fig = plt.figure()
        self.ani = animation.FuncAnimation(fig, self.add_point, interval=self.refresh_interval * 1000)
        self.ax = fig.add_subplot(1, 1, 1)
        plt.gcf().autofmt_xdate()

    def show_graph(self):
        plt.show()

    def add_point(self, _i):
        result = requests.post(self.elasticsearch_endpoint, json=self.query).json()
        x = datetime.now()
        # call get on result while there are dots in self.field
        y = functools.reduce(dict.get, [result] + self.field.split('.'))

        self.xs.append(x)
        self.ys.append(y)
        self.ax.clear()
        self.ax.plot(self.xs, self.ys)

        plt.xticks(rotation=45, ha='right')
        plt.title("Dynamic Plot of %s" % self.field)
        plt.xlabel("Time")
        plt.ylabel(self.field)


@click.command()
@click.option('--query', help='Give a JSON query to filter documents. It can be a file with @path/to/file. Default to all.', default='{"query":{"match_all":{}}}')
@click.option('--index', help='Elasticsearch index (default local-datashare)', default='local-datashare')
@click.option('--refresh-interval', help='Graph refresh interval in seconds (default 5)', default=5)
@click.option('--field', help='Field indicator to display over time (default hits.total.value)', default='hits.total.value')
@click.option('--elasticsearch-url', help='Elasticsearch URL which is used to perform update by query', default='http://elasticsearch:9200')
def graph(**options) -> None:
    GraphRealTime(**options).show_graph()


if __name__ == '__main__':
    graph()



import click
import logging
from tarentula.logger import add_syslog_handler, add_stdout_handler
from tarentula.tag_cleaning_by_query import TagsCleanerByQuery
from tarentula.tagging import Tagger
from tarentula.tagging_by_query import TaggerByQuery
from tarentula.download import Download
from tarentula.export_by_query import ExportByQuery
from tarentula import __version__


def validate_loglevel(ctx, param, value):
    try:
        if isinstance(value, str):
            return getattr(logging, value)
        return int(value)
    except (AttributeError, ValueError):
        raise click.BadParameter('must be a valid log level (CRITICAL, ERROR, WARNING, INFO, DEBUG or NOTSET)')


def validate_progressbar(ctx, param, value):
    # If no value given, we activate the progress bar only when the
    # stdout_loglevel value is higher than INFO (20)
    return value if value is not None else ctx.obj['stdout_loglevel'] > 20


@click.group()
@click.pass_context
@click.version_option(message='v%(version)s', version=__version__)
@click.option('--syslog-address', help='Syslog address', default='localhost')
@click.option('--syslog-port', help='Syslog port', default=514)
@click.option('--syslog-facility', help='Syslog facility', default='local7')
@click.option('--stdout-loglevel', help='Change the default log level for stdout error handler', default='ERROR',
              callback=validate_loglevel)
def cli(ctx, **options):
    # Configure Syslog handler
    add_syslog_handler(options['syslog_address'], options['syslog_port'], options['syslog_facility'])
    add_stdout_handler(options['stdout_loglevel'])
    # Pass all option to context
    ctx.ensure_object(dict)
    ctx.obj.update(options)


@click.command()
@click.option('--datashare-url', help='Datashare URL', default='http://localhost:8080')
@click.option('--datashare-project', help='Datashare project', default='local-datashare')
@click.option('--throttle', help='Request throttling (in ms)', default=0)
@click.option('--cookies', help='Key/value pair to add a cookie to each request to the API. You can separate'
                                'semicolons: key1=val1;key2=val2;...', default='')
@click.option('--apikey', help='Datashare authentication apikey', default=None)
@click.option('--traceback/--no-traceback', help='Display a traceback in case of error', default=False)
@click.option('--progressbar/--no-progressbar', help='Display a progressbar', default=None,
              callback=validate_progressbar)
@click.argument('csv-path', type=click.Path(exists=True))
def tagging(**options):
    # Instantiate a Tagger class with all the options
    tagger = Tagger(**options)
    # Proceed to tagging
    tagger.start()


@click.command()
@click.option('--datashare-project', help='Datashare project', default='local-datashare')
@click.option('--elasticsearch-url', help='Elasticsearch URL which is used to perform update by query',
              default='http://localhost:9200')
@click.option('--throttle', help='Request throttling (in ms)', default=0)
@click.option('--cookies', help='Key/value pair to add a cookie to each request to the API. You can separate'
                                'semicolons: key1=val1;key2=val2;...', default='')
@click.option('--apikey', help='Datashare authentication apikey', default=None)
@click.option('--traceback/--no-traceback', help='Display a traceback in case of error', default=False)
@click.option('--progressbar/--no-progressbar', help='Display a progressbar', default=None,
              callback=validate_progressbar)
@click.option('--wait-for-completion/--no-wait-for-completion', help='Create a Elasticsearch task to perform the update'
                                                                     'asynchronously', default=True)
@click.option('--scroll-size', help='Size of the scroll request that powers the operation.', default=1000)
@click.argument('json-path', type=click.Path(exists=True))
def tagging_by_query(**options):
    # Instantiate a TaggerByQuery class with all the options
    tagger = TaggerByQuery(**options)
    tagger.start()


@click.command()
@click.option('--datashare-project', help='Datashare project', default='local-datashare')
@click.option('--elasticsearch-url', help='Elasticsearch URL which is used to perform update by query',
              default='http://localhost:9200')
@click.option('--cookies', help='Key/value pair to add a cookie to each request to the API. You can separate'
                                'semicolons: key1=val1;key2=val2;...', default='')
@click.option('--apikey', help='Datashare authentication apikey', default=None)
@click.option('--traceback/--no-traceback', help='Display a traceback in case of error', default=False)
@click.option('--wait-for-completion/--no-wait-for-completion', help='Create a Elasticsearch task to perform the update'
                                                                     'asynchronously', default=True)
@click.option('--query', help='Give a JSON query to filter documents that will have their tags cleaned. It can be a'
                              'file with @path/to/file. Default to all.', default=None)
def clean_tags_by_query(**options):
    tagger = TagsCleanerByQuery(**options)
    tagger.start()


@click.command()
@click.option('--datashare-url', help='Datashare URL', default='http://localhost:8080')
@click.option('--datashare-project', help='Datashare project', default='local-datashare')
@click.option('--elasticsearch-url', help='You can additionally pass the Elasticsearch URL in order to use scrolling'
                                          'capabilities of Elasticsearch (useful when dealing with a lot of results)',
              default=None)
@click.option('--query', help='The query string to filter documents', default='*')
@click.option('--destination-directory', help='Directory documents will be downloaded', default='./tmp')
@click.option('--throttle', help='Request throttling (in ms)', default=0)
@click.option('--cookies', help='Key/value pair to add a cookie to each request to the API. You can separate'
                                'semicolons: key1=val1;key2=val2;...', default='')
@click.option('--apikey', help='Datashare authentication apikey', default=None)
@click.option('--path-format', help='Downloaded document path template', default='{id_2b}/{id_4b}/{id}')
@click.option('--scroll', help='Scroll duration', default=None)
@click.option('--source', help='A comma-separated list of field to include in the downloaded document from the index',
              default=None)
@click.option('--once/--not-once', help='Download file only once', default=False)
@click.option('--traceback/--no-traceback', help='Display a traceback in case of error', default=False)
@click.option('--progressbar/--no-progressbar', help='Display a progressbar', default=None,
              callback=validate_progressbar)
@click.option('--raw-file/--no-raw-file', help='Download raw file from Datashare', default=True)
@click.option('--type', help='Type of indexed documents to download', default='Document',
              type=click.Choice(['Document', 'NamedEntity'], case_sensitive=True))
def download(**options):
    # Instantiate a Download class with all the options
    download = Download(**options)
    download.start()


@click.command()
@click.option('--datashare-url', help='Datashare URL', default='http://localhost:8080')
@click.option('--datashare-project', help='Datashare project', default='local-datashare')
@click.option('--elasticsearch-url', help='You can additionally pass the Elasticsearch URL in order to use scrolling'
                                          'capabilities of Elasticsearch (useful when dealing with a lot of results)',
              default=None)
@click.option('--query', help='The query string to filter documents', default='*')
@click.option('--output-file', help='Path to the CSV file', default='tarentula_documents.csv')
@click.option('--throttle', help='Request throttling (in ms)', default=0)
@click.option('--cookies', help='Key/value pair to add a cookie to each request to the API. You can separate'
                                'semicolons: key1=val1;key2=val2;...', default='')
@click.option('--apikey', help='Datashare authentication apikey', default=None)
@click.option('--scroll', help='Scroll duration', default=None)
@click.option('--source', help='A comma-separated list of field to include in the export',
              default='contentType,contentLength:0,extractionDate,path')
@click.option('--once/--not-once', help='Download file only once', default=False)
@click.option('--traceback/--no-traceback', help='Display a traceback in case of error', default=False)
@click.option('--progressbar/--no-progressbar', help='Display a progressbar', default=None,
              callback=validate_progressbar)
@click.option('--type', help='Type of indexed documents to download', default='Document',
              type=click.Choice(['Document', 'NamedEntity'], case_sensitive=True))
def export_by_query(**options):
    # Instantiate an ExportByQuery class with all the options
    export = ExportByQuery(**options)
    export.start()


cli.add_command(tagging)
cli.add_command(download)
cli.add_command(tagging_by_query)
cli.add_command(clean_tags_by_query)
cli.add_command(export_by_query)

if __name__ == '__main__':
    cli()



import csv
import re
import requests
import sys
from time import sleep
from tqdm import tqdm
from http.cookies import SimpleCookie
from requests.exceptions import HTTPError, ConnectionError

from tarentula.logger import logger

DATASHARE_DOCUMENT_ROUTE = re.compile(r'/#/d/[a-zA-Z0-9_-]+/(\w+)(?:/(\w+))?$')


class Tagger:
    def __init__(self,
                 datashare_url: str = 'http://localhost:8080',
                 datashare_project: str = 'local-datashare',
                 throttle: int = 0,
                 csv_path: str = '',
                 cookies: str = '',
                 apikey: str = None,
                 traceback: bool = False,
                 progressbar: bool = True):
        self.datashare_url = datashare_url
        self.datashare_project = datashare_project
        self.cookies_string = cookies
        self.apikey = apikey
        self.throttle = throttle
        self.csv_path = csv_path
        self.traceback = traceback
        self.progressbar = progressbar

    @property
    def no_progressbar(self):
        return not self.progressbar

    @property
    def csv_rows(self):
        with open(self.csv_path, newline='') as csv_file:
            return list(self.sanitize_row(row) for row in csv.DictReader(csv_file))

    @property
    def tags(self):
        return list(dict.fromkeys([row['tag'] for row in self.csv_rows]))

    @property
    def documentIds(self):
        return list(dict.fromkeys([row['documentId'] for row in self.csv_rows]))

    @property
    def tree(self):
        tree = dict()
        for row in self.csv_rows:
            # Extract row values
            tag, document_id, routing = (row['tag'], row['documentId'],
                                         row.get('routing', row['documentId']) or row['documentId'],)
            # Append to an existing dictionary or create one
            tree[document_id] = tree[document_id] if document_id in tree else dict(tags=set(), routing=routing,
                                                                                   document_id=document_id)
            # Tags are added to a set so they are unique
            tree[document_id]['tags'].add(tag)
        return tree

    @property
    def cookies(self):
        cookies = SimpleCookie()
        try:
            cookies.load(self.cookies_string)
            return {key: morsel.value for (key, morsel) in cookies.items()}
        except (TypeError, AttributeError):
            return {}

    def sleep(self):
        sleep(self.throttle / 1000)

    def sanitize_row(self, row):
        if 'documentUrl' in row:
            groups = DATASHARE_DOCUMENT_ROUTE.findall(row['documentUrl'])
            if len(groups) > 0:
                row['documentId'], row['routing'] = groups[0]
        return row

    def leaf_tagging_endpoint(self, leaf):
        document_id, tags, routing = (leaf['document_id'], leaf['tags'], leaf['routing'])
        # @see https://github.com/ICIJ/datashare/wiki/Datashare-API
        url_template = '{datashare_url}/api/{datashare_project}/documents/tags/{document_id}?routing={routing}'
        return url_template.format(
            datashare_url=self.datashare_url,
            datashare_project=self.datashare_project,
            document_id=document_id,
            routing=routing
        )

    def summarize(self):
        summary = 'Adding %s tags to %s documents' % (len(self.tags), len(self.documentIds))
        logger.info(summary)
        return summary

    def start(self):
        for document_id, leaf in tqdm(self.tree.items(), desc=self.summarize(), file=sys.stdout,
                                      disable=self.no_progressbar):
            endpoint_url = self.leaf_tagging_endpoint(leaf)
            for tag in leaf['tags']:
                try:
                    result = requests.put(endpoint_url, json=[tag], cookies=self.cookies,
                                          headers=None if self.apikey is None else
                                          {'Authorization': 'bearer %s' % self.apikey})
                    result.raise_for_status()
                    self.sleep()
                    if result.status_code == requests.codes.ok:
                        logger.info('Tag "%s" already exists on document "%s"' % (tag, document_id,))
                    elif result.status_code == requests.codes.created:
                        logger.info('Added "%s" to document "%s"' % (tag, document_id,))
                except (HTTPError, ConnectionError):
                    logger.warning('Unable to add "%s" to document "%s"' % (tag, document_id), exc_info=self.traceback)



import requests
import uuid

from .test_abstract import TestAbstract


class TestDatashareClient(TestAbstract):

    def test_default_index_creation(self):
        self.assertEqual(requests.get('%s/%s' % (self.elasticsearch_url, self.datashare_project)).status_code,
                         requests.codes.ok)
        self.assertNotEqual(requests.get('%s/%s' % (self.elasticsearch_url, 'unknown-index')).status_code,
                            requests.codes.ok)

    def test_temporary_project_creation(self):
        with self.datashare_client.temporary_project(self.datashare_project) as project:
            result = requests.get('%s/%s' % (self.elasticsearch_url, project))
            self.assertEqual(result.status_code, requests.codes.ok)

    def test_temporary_project_deletion(self):
        project = self.datashare_client.temporary_project(self.datashare_project)
        self.datashare_client.refresh()
        result = requests.get('%s/%s' % (self.elasticsearch_url, project))
        self.assertNotEqual(result.status_code, requests.codes.ok)

    def test_document_creation(self):
        with self.datashare_client.temporary_project(self.datashare_project) as project:
            id = self.datashare_client.index(index=project, document={'foo': 'bar'}, id=str(uuid.uuid4()))
            result = requests.get('%s/%s/_doc/%s' % (self.elasticsearch_url, project, id))
            self.assertEqual(result.status_code, requests.codes.ok)
            self.assertEqual(result.json().get('_source', {}).get('foo'), 'bar')

    def test_document_deletion(self):
        with self.datashare_client.temporary_project(self.datashare_project) as project:
            self.datashare_client.index(index=project, document={'foo': 'bar'}, id='to-delete')
            result = requests.get('%s/%s/_doc/%s' % (self.elasticsearch_url, project, 'to-delete'))
            self.assertTrue(result.json().get('found'))
            self.datashare_client.delete(index=project, id='to-delete')
            result = requests.get('%s/%s/_doc/%s' % (self.elasticsearch_url, project, 'to-delete'))
            self.assertFalse(result.json().get('found'))

    def test_index_deletion(self):
        with self.datashare_client.temporary_project(self.datashare_project) as project:
            result = requests.get('%s/%s' % (self.elasticsearch_url, project))
            self.assertEqual(result.status_code, requests.codes.ok)
            self.datashare_client.delete_index(project)
            result = requests.get('%s/%s' % (self.elasticsearch_url, project))
            self.assertNotEqual(result.status_code, requests.codes.ok)

    def test_query_with_two_docs(self):
        with self.datashare_client.temporary_project(self.datashare_project) as project:
            self.datashare_client.index(index=project, document={'name': 'Atypidae'}, id=str(uuid.uuid4()))
            self.datashare_client.index(index=project, document={'name': 'Migidae'}, id=str(uuid.uuid4()))
            total = self.datashare_client.query(index=project).get('hits', {}).get('total', {}).get('value', None)
            self.assertEqual(total, 2)

    def test_query_with_three_docs(self):
        with self.datashare_client.temporary_project(self.datashare_project) as project:
            self.datashare_client.index(index=project, document={'name': 'Atypidae'}, id=str(uuid.uuid4()))
            self.datashare_client.index(index=project, document={'name': 'Euctenizidae'}, id=str(uuid.uuid4()))
            self.datashare_client.index(index=project, document={'name': 'Migidae'}, id=str(uuid.uuid4()))
            total = self.datashare_client.query(index=project).get('hits', {}).get('total', {}).get('value', None)
            self.assertEqual(total, 3)

    def test_query_has_limited_size(self):
        with self.datashare_client.temporary_project(self.datashare_project) as project:
            self.datashare_client.index(index=project, document={'name': 'Atypidae'}, id=str(uuid.uuid4()))
            self.datashare_client.index(index=project, document={'name': 'Euctenizidae'}, id=str(uuid.uuid4()))
            self.datashare_client.index(index=project, document={'name': 'Migidae'}, id=str(uuid.uuid4()))
            hits = self.datashare_client.query(index=project, size=1).get('hits', {})
            self.assertEqual(hits['total']['value'], 3)
            self.assertEqual(len(hits['hits']), 1)

    def test_query_is_applied_and_get_all_documents(self):
        with self.datashare_client.temporary_project(self.datashare_project) as project:
            self.datashare_client.index(index=project, document={'name': 'Atypidae'}, id=str(uuid.uuid4()))
            self.datashare_client.index(index=project, document={'name': 'Euctenizidae'}, id=str(uuid.uuid4()))
            self.datashare_client.index(index=project, document={'name': 'Migidae'}, id=str(uuid.uuid4()))
            total = self.datashare_client.query(index=project, q='name:*ae').get('hits', {}).get('total', {}).get('value', None)
            self.assertEqual(total, 3)

    def test_query_is_applied_and_get_one_document(self):
        with self.datashare_client.temporary_project(self.datashare_project) as project:
            self.datashare_client.index(index=project, document={'name': 'Atypidae'}, id=str(uuid.uuid4()))
            self.datashare_client.index(index=project, document={'name': 'Euctenizidae'}, id=str(uuid.uuid4()))
            self.datashare_client.index(index=project, document={'name': 'Migidae'}, id=str(uuid.uuid4()))
            total = self.datashare_client.query(index=project, q='name:Migi*').get('hits', {}).get('total', {}).get('value', None)
            self.assertEqual(total, 1)

    def test_get_document(self):
        with self.datashare_client.temporary_project(self.datashare_project) as project:
            self.datashare_client.index(index=project, document={'name': 'Atypidae'}, id='atypidae')
            document = self.datashare_client.document(index=project, id='atypidae')
            self.assertEqual(document.get('_source', {}).get('name'), 'Atypidae')

    def test_query_is_made_over_all_documents(self):
        with self.datashare_client.temporary_project(self.datashare_project) as project:
            self.datashare_client.index(index=project, document={'name': 'Actinopodidae'}, id=str(uuid.uuid4()))
            self.datashare_client.index(index=project, document={'name': 'Antrodiaetidae'}, id=str(uuid.uuid4()))
            self.datashare_client.index(index=project, document={'name': 'Atracidae'}, id=str(uuid.uuid4()))
            self.datashare_client.index(index=project, document={'name': 'Atypidae'}, id=str(uuid.uuid4()))
            self.datashare_client.index(index=project, document={'name': 'Barychelidae'}, id=str(uuid.uuid4()))
            self.datashare_client.index(index=project, document={'name': 'Barychelidae'}, id=str(uuid.uuid4()))
            self.datashare_client.index(index=project, document={'name': 'Ctenizidae'}, id=str(uuid.uuid4()))
            self.datashare_client.index(index=project, document={'name': 'Cyrtaucheniidae'}, id=str(uuid.uuid4()))
            self.datashare_client.index(index=project, document={'name': 'Dipluridae'}, id=str(uuid.uuid4()))

            count = 0
            for document in self.datashare_client.query_all(index=project, q='name:*', size=2):
                count = count + 1
            self.assertEqual(count, 9)

    def test_scan_is_made_over_all_documents(self):
        with self.datashare_client.temporary_project(self.datashare_project) as project:
            self.datashare_client.index(index=project, document={'name': 'Actinopodidae'}, id=str(uuid.uuid4()))
            self.datashare_client.index(index=project, document={'name': 'Antrodiaetidae'}, id=str(uuid.uuid4()))
            self.datashare_client.index(index=project, document={'name': 'Atracidae'}, id=str(uuid.uuid4()))
            self.datashare_client.index(index=project, document={'name': 'Atypidae'}, id=str(uuid.uuid4()))
            self.datashare_client.index(index=project, document={'name': 'Barychelidae'}, id=str(uuid.uuid4()))
            self.datashare_client.index(index=project, document={'name': 'Barychelidae'}, id=str(uuid.uuid4()))
            self.datashare_client.index(index=project, document={'name': 'Ctenizidae'}, id=str(uuid.uuid4()))

            count = 0
            for document in self.datashare_client.scan_all(index=project, q='name:*', size=2):
                count = count + 1
            self.assertEqual(count, 7)



from click.testing import CliRunner

from tarentula.cli import cli
from .test_abstract import TestAbstract, absolute_path


class TestTaggingByQuery(TestAbstract):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.json_tags_path = absolute_path('tests/fixtures/tags-by-content-type.json')

    def test_summary(self):
        runner = CliRunner()
        result = runner.invoke(cli, ['tagging-by-query', '--datashare-project', self.datashare_project, self.json_tags_path])
        self.assertIn('This action will add 8 tag(s)', result.output)

    def test_summary_by_tag(self):
        runner = CliRunner()
        result = runner.invoke(cli, ['tagging-by-query', '--datashare-project', self.datashare_project, self.json_tags_path])
        self.assertIn('Adding "audio-type" tag', result.output)
        self.assertIn('Adding "document-type" tag', result.output)
        self.assertIn('Adding "email-type" tag', result.output)
        self.assertIn('Adding "image-type" tag', result.output)
        self.assertIn('Adding "other-type" tag', result.output)
        self.assertIn('Adding "presentation-type" tag', result.output)
        self.assertIn('Adding "spreadsheet-type" tag', result.output)
        self.assertIn('Adding "video-type" tag', result.output)

    def test_actinopodidae_and_barychelidae_are_audio(self):
        with self.existing_species_documents():
            runner = CliRunner()
            result = runner.invoke(cli, ['tagging-by-query', '--elasticsearch-url', self.elasticsearch_url, '--datashare-project', self.datashare_project, self.json_tags_path])
            actinopodidae = self.datashare_client.document(self.datashare_project, 'l7VnZZEzg2fr960NWWEG')
            tags = actinopodidae.get('_source', {}).get('tags', [])
            self.assertIn('audio-type', tags)
            barychelidae = self.datashare_client.document(self.datashare_project, 'iuL6GUBpO7nKyfSSFaS0')
            tags = barychelidae.get('_source', {}).get('tags', [])
            self.assertIn('audio-type', tags)

    def test_three_documents_are_emails(self):
        with self.existing_species_documents():
            runner = CliRunner()
            result = runner.invoke(cli, ['tagging-by-query', '--elasticsearch-url', self.elasticsearch_url, '--datashare-project', self.datashare_project, self.json_tags_path])
            self.datashare_client.refresh(self.datashare_project)
            emails = self.datashare_client.query(self.datashare_project, q='tags:email-type AND name:*')
            self.assertEqual(emails['hits']['total']['value'], 3)

    def test_actinopodidae_is_tagged_as_audio_once(self):
        with self.existing_species_documents():
            runner = CliRunner()
            runner.invoke(cli, ['tagging-by-query', '--elasticsearch-url', self.elasticsearch_url, '--datashare-project', self.datashare_project, self.json_tags_path])
            runner.invoke(cli, ['tagging-by-query', '--elasticsearch-url', self.elasticsearch_url, '--datashare-project', self.datashare_project, self.json_tags_path])
            actinopodidae = self.datashare_client.document(self.datashare_project, 'l7VnZZEzg2fr960NWWEG')
            tags = actinopodidae.get('_source', {}).get('tags', [])
            self.assertIn('audio-type', tags)
            self.assertEqual(1, len(tags))

    def test_tasks_are_created(self):
        runner = CliRunner()
        result = runner.invoke(cli, ['tagging-by-query', '--elasticsearch-url', self.elasticsearch_url, '--datashare-project', self.datashare_project, self.json_tags_path, '--no-wait-for-completion'])
        self.assertEqual(result.output.count('task created'), 8)
        self.assertEqual(result.output.count('documents updated in'), 0)

    def test_tasks_are_not_created(self):
        runner = CliRunner()
        result = runner.invoke(cli, ['tagging-by-query', '--elasticsearch-url', self.elasticsearch_url, '--datashare-project', self.datashare_project, self.json_tags_path, '--wait-for-completion'])
        self.assertEqual(result.output.count('task created'), 0)
        self.assertEqual(result.output.count('documents updated in'), 8)

    def test_progressbar_is_not_created(self):
        runner = CliRunner()
        result = runner.invoke(cli, ['tagging-by-query', '--elasticsearch-url', self.elasticsearch_url, '--datashare-project', self.datashare_project, self.json_tags_path, '--no-progressbar'])
        self.assertNotIn('This action will add 8 tag(s)', result.output)

    def test_logs_are_printed_to_stdout(self):
        runner = CliRunner()
        result = runner.invoke(cli, ['--stdout-loglevel', 'INFO', 'tagging-by-query', '--elasticsearch-url', self.elasticsearch_url, '--datashare-project', self.datashare_project, self.json_tags_path])
        self.assertNotIn('This action will add 8 tag(s)', result.output)
        self.assertEqual(result.output.count('Documents tagged with'), 16)  # logs + stdout



import csv

from click.testing import CliRunner
from datetime import datetime
from os.path import join
from tempfile import TemporaryDirectory

from tarentula.cli import cli
from .test_abstract import TestAbstract


class TestExportByQuery(TestAbstract):
    def tearDown(self):
        super().tearDown()

    def test_csv_file(self):
        with self.existing_species_documents(), TemporaryDirectory() as tmp:
            output_file = join(tmp, 'output.csv')
            runner = CliRunner()
            runner.invoke(cli, ['export-by-query', '--datashare-url', self.datashare_url, '--elasticsearch-url',
                                self.elasticsearch_url, '--datashare-project', self.datashare_project, '--query',
                                'Actinopodidae OR Antrodiaetidae', '--output-file', output_file])
            with open(output_file, newline='') as csv_file:
                csv_reader = csv.DictReader(csv_file)

                # First row
                row = next(csv_reader)
                self.assertEqual(row['query'], 'Actinopodidae OR Antrodiaetidae')
                self.assertEqual(row['documentUrl'],
                                 'http://localhost:8080/#/d/test-datashare/DWLOskax28jPQ2CjFrCo/l7VnZZEzg2fr960NWWEG')
                self.assertEqual(row['documentId'], 'DWLOskax28jPQ2CjFrCo')
                self.assertEqual(row['rootId'], 'l7VnZZEzg2fr960NWWEG')
                self.assertEqual(row['contentType'], 'audio/vnd.wave')
                self.assertEqual(row['contentLength'], '0')
                self.assertEqual(row['path'], '')
                datetime_object = datetime.strptime(row['extractionDate'], '%Y-%m-%dT%H:%M:%S.%fZ')
                self.assertIsInstance(datetime_object, datetime)
                self.assertEqual(row['documentNumber'], '0')

                # Second row
                row = next(csv_reader)
                self.assertEqual(row['query'], 'Actinopodidae OR Antrodiaetidae')
                self.assertEqual(row['documentUrl'],
                                 'http://localhost:8080/#/d/test-datashare/l7VnZZEzg2fr960NWWEG/l7VnZZEzg2fr960NWWEG')
                self.assertEqual(row['documentId'], 'l7VnZZEzg2fr960NWWEG')
                self.assertEqual(row['rootId'], 'l7VnZZEzg2fr960NWWEG')
                self.assertEqual(row['contentType'], 'audio/mpeg')
                self.assertEqual(row['contentLength'], '25')
                self.assertEqual(row['path'], '/path/to/file.txt')
                datetime_object = datetime.strptime(row['extractionDate'], '%Y-%m-%dT%H:%M:%S.%fZ')
                self.assertIsInstance(datetime_object, datetime)
                self.assertEqual(row['documentNumber'], '1')



import json
from os.path import join
from tempfile import TemporaryDirectory

from click.testing import CliRunner

from .test_abstract import TestAbstract
from tarentula.cli import cli


def load_json_file(path):
    return json.loads(open(path).read())


class TestDownload(TestAbstract):

    def tearDown(self):
        super().tearDown()

    def test_summary(self):
        with self.existing_species_documents():
            runner = CliRunner()
            result = runner.invoke(cli, ['download', '--datashare-url', self.datashare_url, '--elasticsearch-url', self.elasticsearch_url, '--datashare-project', self.datashare_project, '--no-raw-file', '--query', 'name:*'])
            self.assertIn('Downloading 20 document(s)', result.output)

    def test_summary_with_scroll(self):
        with self.existing_species_documents():
            runner = CliRunner()
            result = runner.invoke(cli, ['download', '--datashare-url', self.datashare_url, '--elasticsearch-url', self.elasticsearch_url, '--datashare-project', self.datashare_project, '--no-raw-file', '--query', 'name:*', '--scroll', '1m'])
            self.assertIn('Downloading 20 document(s)', result.output)

    def test_summary_with_wildcard(self):
        with self.existing_species_documents():
            runner = CliRunner()
            result = runner.invoke(cli, ['download', '--datashare-url', self.datashare_url, '--elasticsearch-url', self.elasticsearch_url, '--datashare-project', self.datashare_project, '--no-raw-file', '--query', 'name:*dae'])
            self.assertIn('Downloading 20 document(s)', result.output)

    def test_summary_with_wildcard_sta(self):
        with self.existing_species_documents():
            runner = CliRunner()
            result = runner.invoke(cli, ['download', '--datashare-url', self.datashare_url, '--elasticsearch-url', self.elasticsearch_url, '--datashare-project', self.datashare_project, '--no-raw-file', '--query', 'name:*dae'])
            self.assertIn('Downloading 20 document(s)', result.output)

    def test_meta_is_downloaded_for_actinopodidae(self):
        with self.existing_species_documents(), TemporaryDirectory() as tmp:
            runner = CliRunner()
            runner.invoke(cli, ['download', '--datashare-url', self.datashare_url, '--elasticsearch-url', self.elasticsearch_url, '--datashare-project',
                                self.datashare_project, '--no-raw-file', '--destination-directory', tmp, '--query', 'name:Actinopodidae'])
            json_file = load_json_file(join(tmp, 'l7/Vn/l7VnZZEzg2fr960NWWEG.json'))
            self.assertEqual(json_file['_id'], 'l7VnZZEzg2fr960NWWEG')

    def test_meta_is_downloaded_for_ctenizidae(self):
        with self.existing_species_documents(), TemporaryDirectory() as tmp:
            runner = CliRunner()
            runner.invoke(cli, ['download', '--datashare-url', self.datashare_url, '--elasticsearch-url', self.elasticsearch_url, '--datashare-project',
                                self.datashare_project, '--no-raw-file', '--destination-directory', tmp, '--query', 'name:Ctenizidae'])
            json_file = load_json_file(join(tmp, 'Bm/ov/BmovvXBisWtyyx6o9cuG.json'))
            self.assertEqual(json_file['_id'], 'BmovvXBisWtyyx6o9cuG')

    def test_meta_is_downloaded_for_idiopidae_with_default_properties(self):
        with self.existing_species_documents(), TemporaryDirectory() as tmp:
            runner = CliRunner()
            runner.invoke(cli, ['download', '--datashare-url', self.datashare_url, '--elasticsearch-url', self.elasticsearch_url, '--datashare-project',
                                self.datashare_project, '--no-raw-file', '--destination-directory', tmp, '--query', 'name:Idiopidae'])
            json_file = load_json_file(join(tmp, 'Dz/LO/DzLOskax28jPQ2CjFrCo.json'))
            self.assertIn('_id', json_file)
            self.assertIn('_source', json_file)
            self.assertNotIn('name', json_file['_source'])

    def test_meta_is_downloaded_for_idiopidae_with_extra_properties(self):
        with self.existing_species_documents(), TemporaryDirectory() as tmp:
            runner = CliRunner()
            runner.invoke(cli, ['download', '--datashare-url', self.datashare_url, '--elasticsearch-url', self.elasticsearch_url, '--datashare-project', self.datashare_project, '--destination-directory', tmp, '--no-raw-file', '--query', 'name:Idiopidae', '--source', 'name'])
            json = load_json_file(join(tmp, 'Dz/LO/DzLOskax28jPQ2CjFrCo.json'))
            self.assertIn('_id', json)
            self.assertIn('_source', json)
            self.assertIn('name', json['_source'])



import json
import shutil
import sys
from os import makedirs
from os.path import join, dirname, basename, exists
from time import sleep

from requests.exceptions import HTTPError, ConnectionError
from tqdm import tqdm
from urllib3.exceptions import ProtocolError

from tarentula.datashare_client import DatashareClient
from tarentula.logger import logger


class Download:
    def __init__(self,
                 datashare_url: str = 'http://localhost:8080',
                 datashare_project: str = 'local-datashare',
                 destination_directory: str = './tmp',
                 query: str = '*',
                 throttle: int = 0,
                 cookies: str = '',
                 apikey: str = None,
                 elasticsearch_url: str = None,
                 path_format: str = '{id_2b}/{id_4b}/{id}',
                 scroll: str = None,
                 once: bool = False,
                 traceback: bool = False,
                 progressbar: bool = True,
                 raw_file: bool = True,
                 source: str = None,
                 type: str = 'Document'):
        self.datashare_url = datashare_url
        self.datashare_project = datashare_project
        self.query = query
        self.destination_directory = destination_directory
        self.throttle = throttle
        self.cookies_string = cookies
        self.apikey = apikey
        self.path_format = path_format
        self.once = once
        self.traceback = traceback
        self.progressbar = progressbar
        self.raw_file = raw_file
        self.source = source
        self.scroll = scroll
        self.type = type
        try:
            self.datashare_client = DatashareClient(datashare_url,
                                                    elasticsearch_url,
                                                    datashare_project,
                                                    cookies,
                                                    apikey)
        except (ConnectionRefusedError, ConnectionError):
            logger.critical('Unable to connect to Datashare', exc_info=self.traceback)
            exit()

    @property
    def query_body(self):
        if self.query.startswith('@'):
            return self.query_body_from_file
        else:
            return self.query_body_from_string

    @property
    def query_body_from_string(self):
        return {
            "query": {
                "bool": {
                    "must": [
                        {
                            "match": {
                                "type": self.type
                            }
                        },
                        {
                            "query_string": {
                                "query": self.query
                            }
                        }
                    ]
                }
            }
        }

    @property
    def query_body_from_file(self):
        with open(self.query[1:]) as json_file:
            query_body = json.load(json_file)
        return query_body

    @property
    def no_progressbar(self):
        return not self.progressbar

    def sleep(self):
        sleep(self.throttle / 1000)

    def document_file_options(self, document):
        return {
            "id": document.get('_id'),
            "id_2b": document.get('_id')[0:2],
            "id_4b": document.get('_id')[2:4],
            "project": self.datashare_project,
            "basename": basename(document.get('_source', {}).get("path", '')),
            "parentDocument": document.get('_source', {}).get('parentDocument', None)
        }

    def raw_file_path(self, document, parents=True):
        formatted_path = self.path_format.format(**self.document_file_options(document))
        file_path = join(self.destination_directory, formatted_path)
        if parents:
            parents_path = dirname(file_path)
            makedirs(parents_path, exist_ok=True)
        return file_path

    def indexed_document_path(self, document, parents=True):
        formatted_path = self.path_format.format(**self.document_file_options(document))
        formatted_path = '.'.join((formatted_path, 'json'))
        file_path = join(self.destination_directory, formatted_path)
        if parents:
            parents_path = dirname(file_path)
            makedirs(parents_path, exist_ok=True)
        return file_path

    def count_matches(self):
        index = self.datashare_project
        return self.datashare_client.count(index=index, query=self.query_body).get('count')

    def log_matches(self):
        index = self.datashare_project
        count = self.count_matches()
        logger.info('%s matching document(s) in %s' % (count, index))
        return count

    def scan_or_query_all(self):
        index = self.datashare_project
        source = ["path", "parentDocument", "type"] + str(self.source).split(',')
        if self.scroll is None:
            logger.info('Searching document(s) metadata in %s' % index)
            return self.datashare_client.query_all(index=index, query=self.query_body, source=source)
        else:
            logger.info('Scrolling over document(s) metadata in %s' % index)
            return self.datashare_client.scan_all(index=index, query=self.query_body, source=source, scroll=self.scroll)

    def download_raw_file(self, document):
        id = document.get('_id')
        routing = document.get('_routing', id)
        # Skip raw file
        if not self.raw_file:
            return
        # Skip existing
        if self.once and self.raw_file_exists(document):
            return logger.info('Skipping existing document %s' % document.get('_id'))
        # Skip non-downloadable file
        if document.get('_source', {}).get('type', None) != 'Document':
            return logger.warning('Not a raw document. Skipping %s' % id)
        logger.info('Downloading raw file %s' % id)
        document_file_stream = self.datashare_client.download(self.datashare_project, id, routing)
        document_file_stream.raw.decode_content = True
        document_file_stream.raise_for_status()
        self.save_raw_file(document, document_file_stream)

    def raw_file_exists(self, document):
        raw_file_path = self.raw_file_path(document)
        return exists(raw_file_path)

    def save_raw_file(self, document, document_file_stream):
        file_path = self.raw_file_path(document)
        with open(file_path, 'wb') as file:
            shutil.copyfileobj(document_file_stream.raw, file)

    def save_indexed_document(self, indexed_document):
        file_path = self.indexed_document_path(indexed_document)
        with open(file_path, 'w') as file:
            json.dump(indexed_document, file)

    def start(self):
        count = self.log_matches()
        try:
            documents = self.scan_or_query_all()
            pbar = tqdm(documents, total=count, desc="Downloading %s document(s)" % count, file=sys.stdout,
                        disable=self.no_progressbar)
            for document in pbar:
                try:
                    self.download_raw_file(document)
                    self.save_indexed_document(document)
                    logger.info('Processed document %s' % document.get('_id'))
                    self.sleep()
                except HTTPError:
                    logger.error('Unable to download document %s' % document.get('_id'), exc_info=self.traceback)
        except ProtocolError:
            logger.error('Exception while downloading documents', exc_info=self.traceback)



import json
import os

from contextlib import contextmanager
from os.path import dirname
from unittest import TestCase

from tarentula.datashare_client import DatashareClient


def absolute_path(relative_path):
    return os.path.join(os.path.abspath(dirname(dirname(__file__))), relative_path)


class TestAbstract(TestCase):
    datashare_client = None
    datashare_project = 'test-datashare'

    @classmethod
    def setUpClass(cls):
        cls.elasticsearch_url = os.environ.get('TEST_ELASTICSEARCH_URL', 'http://elasticsearch:9200')
        cls.datashare_url = os.environ.get('TEST_DATASHARE_URL', 'http://localhost:8080')
        cls.datashare_client = DatashareClient(cls.datashare_url, cls.elasticsearch_url)
        cls.species_path = absolute_path('tests/fixtures/species.json')
        cls.datashare_client.create(cls.datashare_project)

    @classmethod
    def tearDownClass(cls):
        cls.datashare_client.delete_index(cls.datashare_project)

    def index_documents(self, documents=None):
        if documents is None:
            documents = []
        for document in documents:
            id = document.get('_id', None)
            routing = document.get('_routing', None)
            self.datashare_client.index(index=self.datashare_project, document=document, id=id, routing=routing)
        self.datashare_client.refresh(index=self.datashare_project)

    def delete_documents(self, documents=None):
        if documents is None:
            documents = []
        for document in documents:
            self.datashare_client.delete(index=self.datashare_project, id=document['_id'])
        self.datashare_client.refresh(index=self.datashare_project)

    @contextmanager
    def existing_species_documents(self):
        with open(self.species_path, 'r') as species_file:
            species = json.loads(species_file.read())
            self.index_documents(species)
            try:
                yield species
            finally:
                self.delete_documents(species)



import csv
import re
import responses

from click.testing import CliRunner
from contextlib import contextmanager
from tempfile import NamedTemporaryFile

from .test_abstract import TestAbstract
from tarentula.cli import cli
from tarentula.tagging import Tagger


class TestTagging(TestAbstract):

    @contextmanager
    def mock_tagging_endpoint(self):
        with responses.RequestsMock() as resp:
            tagging_endpoint_re = r"^%s\/api/%s/documents/tags/" % (self.datashare_url, self.datashare_project)
            resp.add(responses.PUT, re.compile(tagging_endpoint_re), body='', status=201)
            yield resp

    def test_summary(self):
        with NamedTemporaryFile() as file:
            file.write(b'tag,documentId,routing\n'
                       b'Actinopodidae,l7VnZZEzg2fr960NWWEG,l7VnZZEzg2fr960NWWEG\n'
                       b'Antrodiaetidae,DWLOskax28jPQ2CjFrCo')
            file.flush()
            file.seek(0)
            with self.mock_tagging_endpoint():
                runner = CliRunner()
                result = runner.invoke(cli, ['tagging', '--datashare-url', self.datashare_url, '--datashare-project',
                                             self.datashare_project, file.name])
                self.assertIn('Adding 2 tags to 2 documents', result.output)

    def test_progression(self):
        with NamedTemporaryFile() as file:
            file.write(b'tag,documentId,routing\n'
                       b'Actinopodidae,l7VnZZEzg2fr960NWWEG,l7VnZZEzg2fr960NWWEG\n'
                       b'Antrodiaetidae,DWLOskax28jPQ2CjFrCo')
            file.flush()
            file.seek(0)
            with self.mock_tagging_endpoint():
                runner = CliRunner()
                result = runner.invoke(cli, ['--stdout-loglevel', 'INFO', 'tagging', '--datashare-url',
                                             self.datashare_url, '--datashare-project', self.datashare_project,
                                             file.name])
                self.assertIn('Added "Actinopodidae" to document "l7VnZZEzg2fr960NWWEG"', result.output)
                self.assertIn('Added "Antrodiaetidae" to document "DWLOskax28jPQ2CjFrCo"', result.output)
                self.assertTrue(result.exit_code == 0)

    def test_http_tagging_requests(self):
        with NamedTemporaryFile() as file:
            file.write(b'tag,documentId,routing\n'
                       b'Actinopodidae,l7VnZZEzg2fr960NWWEG,l7VnZZEzg2fr960NWWEG\n'
                       b'Antrodiaetidae,DWLOskax28jPQ2CjFrCo')
            file.flush()
            file.seek(0)
            with self.mock_tagging_endpoint() as resp:
                runner = CliRunner()
                runner.invoke(cli, ['tagging', '--datashare-url', self.datashare_url, '--datashare-project',
                                    self.datashare_project, file.name])
                self.assertTrue(len(resp.calls) == 2)

    def test_routing_is_correct(self):
        with NamedTemporaryFile() as file:
            file.write(b'tag,documentId,routing\n'
                       b'Actinopodidae,l7VnZZEzg2fr960NWWEG,l7VnZZEzg2fr960NWWEG\n'
                       b'Atracidae,6VE7cVlWszkUd94XeuSd,vZJQpKQYhcI577gJR0aN')
            file.flush()
            file.seek(0)
            tagger = Tagger(self.datashare_url, self.datashare_project, 0, file.name, progressbar=False)
            self.assertEqual(tagger.tree['l7VnZZEzg2fr960NWWEG']['routing'], 'l7VnZZEzg2fr960NWWEG')
            self.assertEqual(tagger.tree['6VE7cVlWszkUd94XeuSd']['routing'], 'vZJQpKQYhcI577gJR0aN')

    def test_routing_uses_fallback(self):
        with NamedTemporaryFile() as file:
            file.write(b'tag,documentId,routing\n'
                       b'Antrodiaetidae,DWLOskax28jPQ2CjFrCo')
            file.flush()
            file.seek(0)
            tagger = Tagger(self.datashare_url, self.datashare_project, 0, file.name, progressbar=False)
            self.assertEqual(tagger.tree['DWLOskax28jPQ2CjFrCo']['routing'], 'DWLOskax28jPQ2CjFrCo')

    def test_document_is_in_tree(self):
        with NamedTemporaryFile() as file:
            file.write(b'tag,documentId,routing\n'
                       b'Actinopodidae,l7VnZZEzg2fr960NWWEG,l7VnZZEzg2fr960NWWEG')
            file.flush()
            file.seek(0)
            tagger = Tagger(self.datashare_url, self.datashare_project, 0, file.name, progressbar=False)
            self.assertIn('l7VnZZEzg2fr960NWWEG', tagger.tree)

    def test_tags_are_in_tree(self):
        with NamedTemporaryFile() as file:
            file.write(b'tag,documentId,routing\n'
                       b'Antrodiaetidae,DWLOskax28jPQ2CjFrCo\n'
                       b'Idiopidae,DWLOskax28jPQ2CjFrCo,DWLOskax28jPQ2CjFrCo')
            file.flush()
            file.seek(0)
            tagger = Tagger(self.datashare_url, self.datashare_project, 0, file.name, progressbar=False)
            self.assertIn('Antrodiaetidae', tagger.tree['DWLOskax28jPQ2CjFrCo']['tags'])
            self.assertIn('Idiopidae', tagger.tree['DWLOskax28jPQ2CjFrCo']['tags'])

    def test_document_has_two_tags(self):
        with NamedTemporaryFile() as file:
            file.write(b'tag,documentId,routing\n'
                       b'Antrodiaetidae,DWLOskax28jPQ2CjFrCo\n'
                       b'Idiopidae,DWLOskax28jPQ2CjFrCo,DWLOskax28jPQ2CjFrCo')
            file.flush()
            file.seek(0)
            tagger = Tagger(self.datashare_url, self.datashare_project, 0, file.name)
            self.assertEqual(len(tagger.tree['DWLOskax28jPQ2CjFrCo']['tags']), 2)

    def test_tags_are_in_tree_with_url(self):
        with NamedTemporaryFile() as file:
            file.write(b'tag,documentUrl\n'
                       b'Antrodiaetidae,http://localhost:8080/#/d/local-datashare/DWLOskax28jPQ2CjFrCo\n'
                       b'Idiopidae,http://localhost:8080/#/d/local-datashare/DWLOskax28jPQ2CjFrCo/DWLOskax28jPQ2CjFrCo')
            file.flush()
            file.seek(0)
            tagger = Tagger(self.datashare_url, self.datashare_project, 0, file.name)
            self.assertIn('Antrodiaetidae', tagger.tree['DWLOskax28jPQ2CjFrCo']['tags'])
            self.assertIn('Idiopidae', tagger.tree['DWLOskax28jPQ2CjFrCo']['tags'])

    def test_routing_is_correct_with_url(self):
        with NamedTemporaryFile() as file:
            file.write(b'tag,documentUrl\n'
                       b'Actinopodidae,http://localhost:8080/#/d/local-datashare/l7VnZZEzg2fr960NWWEG/'
                       b'l7VnZZEzg2fr960NWWEG\n'
                       b'Atracidae,http://localhost:8080/#/d/local-datashare/6VE7cVlWszkUd94XeuSd/vZJQpKQYhcI577gJR0aN')
            file.flush()
            file.seek(0)
            tagger = Tagger(self.datashare_url, self.datashare_project, 0, file.name)
            self.assertEqual(tagger.tree['l7VnZZEzg2fr960NWWEG']['routing'], 'l7VnZZEzg2fr960NWWEG')
            self.assertEqual(tagger.tree['6VE7cVlWszkUd94XeuSd']['routing'], 'vZJQpKQYhcI577gJR0aN')

    def test_tags_are_all_created(self):
        with NamedTemporaryFile() as file:
            file.write(b'tag,documentId,routing\n'
                       b'Actinopodidae,l7VnZZEzg2fr960NWWEG,l7VnZZEzg2fr960NWWEG\n'
                       b'Antrodiaetidae,DWLOskax28jPQ2CjFrCo')
            file.flush()
            file.seek(0)
            with self.datashare_client.temporary_project(self.datashare_project) as project:
                tagger = Tagger(self.datashare_url, project, 0, file.name, progressbar=False)
                # Ensure there is no documents yet
                self.assertEqual(self.datashare_client.query(index=project, size=0).get('hits', {}).get('total', {}).get('value', None), 0)
                # Create all the docs
                for document_id, leaf in tagger.tree.items():
                    self.datashare_client.index(project, {'tags': []}, document_id, leaf['routing'])
                # Ensure the docs exists
                self.assertEqual(self.datashare_client.query(index=project, size=0).get('hits', {}).get('total', {}).get('value', None), 2)
                # Ensure the docs are not tagged yet
                self.assertEqual(self.datashare_client.query(index=project, size=0, q='tags:*')
                                 .get('hits', {}).get('total', {}).get('value', None), 0)
                # Tag them all!
                tagger.start()
                # Refresh the index
                self.datashare_client.refresh(project)
                # Ensure the docs have been tagged
                self.assertEqual(self.datashare_client.query(index=project, size=0, q='tags:*')
                                 .get('hits', {}).get('total', {}).get('value', None), 2)

    def test_tag_is_correct(self):
        with self.datashare_client.temporary_project(self.datashare_project) as project:
            # Create the document
            self.datashare_client.index(index=project,  document={'name': 'Atypidae', 'tags': []}, id='atypidae')
            runner = CliRunner()
            with runner.isolated_filesystem():
                # Create a small CSV with just one tag
                with open('tags.csv', 'w') as file:
                    writer = csv.writer(file)
                    writer.writerows([['tag', 'documentId'], ['mygalomorph', 'atypidae']])
                runner.invoke(cli, ['tagging', '--datashare-url', self.datashare_url, '--datashare-project', project,
                                    'tags.csv'])
            # Get the document from Elasticsearch
            document = self.datashare_client.document(index=project, id='atypidae')
            self.assertIn('mygalomorph', document.get('_source', {}).get('tags', []))

    def test_tags_are_correct(self):
        with self.datashare_client.temporary_project(self.datashare_project) as project:
            # Create the document
            self.datashare_client.index(index=project, document={'name': 'Atypidae', 'tags': []}, id='atypidae')
            runner = CliRunner()
            with runner.isolated_filesystem():
                # Create a small CSV with just one tag
                with open('tags.csv', 'w') as file:
                    writer = csv.writer(file)
                    writer.writerows([['tag', 'documentId'], ['mygalomorph', 'atypidae'], ['spider', 'atypidae']])
                runner.invoke(cli, ['tagging', '--datashare-url', self.datashare_url, '--datashare-project',
                                    project, 'tags.csv'])
            # Get the document from Elasticsearch
            document = self.datashare_client.document(index=project, id='atypidae')
            self.assertIn('mygalomorph', document.get('_source').get('tags', []))
            self.assertIn('spider', document.get('_source').get('tags', []))

    def test_cookies_are_parsed_by_tagger(self):
        with NamedTemporaryFile() as file:
            cookies = 'session=mygalomorph;age=14'
            tagger = Tagger(self.datashare_url, self.datashare_project, 0, file.name, cookies, progressbar=False)
            self.assertEqual(tagger.cookies['session'], 'mygalomorph')
            self.assertEqual(tagger.cookies['age'], '14')
            self.assertEqual(len(tagger.cookies.keys()), 2)

    def test_multiple_cookies_are_parsed_by_tagger(self):
        with NamedTemporaryFile() as file:
            cookies = 'session=mygalomorph'
            tagger = Tagger(self.datashare_url, self.datashare_project, 0, file.name, cookies, progressbar=False)
            self.assertEqual(tagger.cookies['session'], 'mygalomorph')
            self.assertEqual(len(tagger.cookies.keys()), 1)

    def test_session_cookies_are_parsed_by_tagger(self):
        with NamedTemporaryFile() as file:
            cookies = r'_ds_session_id="{\"login\":\"\",\"roles\":[],\"sessionId\":\"dq18s0kj08dq10skLYGSu8SFVsg\",' \
                      r'\"redirectAfterLogin\":\"/\"}"'
            tagger = Tagger(self.datashare_url, self.datashare_project, 0, file.name, cookies, progressbar=False)
            self.assertEqual(tagger.cookies['_ds_session_id'], r'{"login":"","roles":[],"sessionId":'
                                                               r'"dq18s0kj08dq10skLYGSu8SFVsg",'
                                                               r'"redirectAfterLogin":"/"}')
            self.assertEqual(len(tagger.cookies.keys()), 1)

    def test_cookies_are_sent_while_tagging(self):
        with NamedTemporaryFile() as file:
            file.write(b'tag,documentId,routing\n'
                       b'Actinopodidae,l7VnZZEzg2fr960NWWEG,l7VnZZEzg2fr960NWWEG\n'
                       b'Antrodiaetidae,DWLOskax28jPQ2CjFrCo')
            file.flush()
            file.seek(0)
            with self.mock_tagging_endpoint() as resp:
                cookies = 'session=mygalomorph'
                Tagger(self.datashare_url, self.datashare_project, 0, file.name, cookies, progressbar=False).start()
                self.assertEqual(resp.calls[0].request.headers['Cookie'], cookies)
                self.assertEqual(resp.calls[1].request.headers['Cookie'], cookies)

    def test_cookies_are_NOT_sent_while_tagging(self):
        with NamedTemporaryFile() as file:
            file.write(b'tag,documentId,routing\n'
                       b'Actinopodidae,l7VnZZEzg2fr960NWWEG,l7VnZZEzg2fr960NWWEG\n'
                       b'Antrodiaetidae,DWLOskax28jPQ2CjFrCo')
            file.flush()
            file.seek(0)
            with self.mock_tagging_endpoint() as resp:
                Tagger(self.datashare_url, self.datashare_project, 0, file.name, progressbar=False).start()
                self.assertNotIn('Cookie', resp.calls[0].request.headers)
                self.assertNotIn('Cookie', resp.calls[1].request.headers)

    def test_cookies_are_sent_while_tagging_with_the_cli(self):
        with NamedTemporaryFile() as file:
            file.write(b'tag,documentId,routing\n'
                       b'Actinopodidae,l7VnZZEzg2fr960NWWEG,l7VnZZEzg2fr960NWWEG\n'
                       b'Antrodiaetidae,DWLOskax28jPQ2CjFrCo')
            file.flush()
            file.seek(0)
            with self.mock_tagging_endpoint() as resp:
                cookies = r'_ds_session_id="{\"login\":\"\",\"roles\":[],\"sessionId\":' \
                          r'\"dq18s0kj08dq10skLYGSu8SFVsg\",\"redirectAfterLogin\":\"/\"}"'
                cookies_not_escaped = '_ds_session_id={"login":"","roles":[],"sessionId":' \
                                      '"dq18s0kj08dq10skLYGSu8SFVsg","redirectAfterLogin":"/"}'
                runner = CliRunner()
                runner.invoke(cli, ['tagging', '--datashare-url', self.datashare_url, '--datashare-project',
                                    self.datashare_project, '--cookies', cookies, file.name])
                self.assertEqual(resp.calls[0].request.headers['Cookie'], cookies_not_escaped)
                self.assertEqual(resp.calls[1].request.headers['Cookie'], cookies_not_escaped)



import json
from http.cookies import SimpleCookie

import requests

from tarentula.logger import logger


class TagsCleanerByQuery:
    def __init__(self,
                 datashare_project: str = 'local-datashare',
                 elasticsearch_url: str = 'http://localhost:9200',
                 cookies: str = '',
                 apikey: str = None,
                 traceback: bool = False,
                 wait_for_completion: bool = True,
                 query: str = None):
        if query is None:
            self.query = {"query": {"match_all": {}}}
        elif query.startswith('@'):
            with open(query[1:]) as f:
                self.query = json.loads(f.read())
        else:
            self.query = json.loads(query)
        self.datashare_project = datashare_project
        self.elasticsearch_url = elasticsearch_url
        self.cookies_string = cookies
        self.apikey = apikey
        self.traceback = traceback
        self.wait_for_completion = wait_for_completion

    @property
    def cookies(self):
        cookies = SimpleCookie()
        try:
            cookies.load(self.cookies_string)
            return {key: morsel.value for (key, morsel) in cookies.items()}
        except (TypeError, AttributeError):
            return {}

    @property
    def tagging_by_query_endpoint(self):
        url_template = '{elasticsearch_url}/{datashare_project}/_update_by_query?conflicts=proceed'
        return url_template.format(elasticsearch_url=self.elasticsearch_url, datashare_project=self.datashare_project)

    def start(self):
        logger.info("This action will remove all tags for documents matching query")
        script = {"script": {"source": "ctx._source['tags'] = []"}}
        params = {"wait_for_completion": str(self.wait_for_completion).lower()}
        result = requests.post(self.tagging_by_query_endpoint, params=params, json={**script, **self.query},
                               cookies=self.cookies,
                               headers=None if self.apikey is None else {'Authorization': 'bearer %s' % self.apikey})
        result.raise_for_status()
        if self.wait_for_completion:
            logger.info('updated %s documents' % result.json()['updated'])
        else:
            logger.info('task created: [%s]' % result.json()['task'])
        return result



import json
import re
import responses

from click.testing import CliRunner
from contextlib import contextmanager
from tempfile import NamedTemporaryFile

from tarentula.cli import cli
from .test_abstract import TestAbstract


class TestApikey(TestAbstract):

    @contextmanager
    def mock_tagging_endpoint(self):
        with responses.RequestsMock() as resp:
            tagging_endpoint_re = r"^%s\/api/%s/documents/tags/" % (self.datashare_url, self.datashare_project)
            resp.add(responses.PUT, re.compile(tagging_endpoint_re), body='', status=201)
            yield resp

    @contextmanager
    def mock_tagging_by_query_endpoint(self):
        with responses.RequestsMock() as resp:
            tagging_endpoint_re = r"^%s\/%s\/_update_by_query" % (self.elasticsearch_url, self.datashare_project)
            resp.add(responses.POST, re.compile(tagging_endpoint_re), body='', status=201)
            yield resp

    @contextmanager
    def mock_download_endpoint(self):
        with responses.RequestsMock() as resp:
            tagging_endpoint_re = r"^%s\/api/%s/documents/src" % (self.datashare_url, self.datashare_project)
            resp.add(responses.GET, re.compile(tagging_endpoint_re), body='', status=201)
            resp.add_passthru(self.datashare_url)
            resp.add_passthru(self.elasticsearch_url)
            yield resp

    def test_apikey_header_is_NOT_sent_while_tagging_with_cli(self):
        with NamedTemporaryFile() as file:
            file.write(b'tag,documentId,routing\ndocumentTag,documentId,documentRouting')
            file.flush()
            file.seek(0)
            with self.mock_tagging_endpoint() as resp:
                runner = CliRunner()
                runner.invoke(cli, ['tagging', '--datashare-url', self.datashare_url, '--datashare-project',
                                    self.datashare_project, file.name])
                self.assertIsNone(resp.calls[0].request.headers.get('Authorization'))

    def test_apikey_header_is_sent_while_tagging_with_cli(self):
        with NamedTemporaryFile() as file:
            file.write(b'tag,documentId,routing\ndocumentTag,documentId,documentRouting')
            file.flush()
            file.seek(0)
            with self.mock_tagging_endpoint() as resp:
                runner = CliRunner()
                runner.invoke(cli, ['tagging', '--datashare-url', self.datashare_url, '--datashare-project',
                                    self.datashare_project, '--apikey',
                                    'my_api_key', file.name])
                self.assertEqual(resp.calls[0].request.headers['Authorization'], 'bearer my_api_key')

    def test_apikey_header_is_NOT_sent_while_tagging_by_query_with_cli(self):
        with NamedTemporaryFile(mode='w+') as file:
            json.dump({"tag-name": {"query": {"match_all": {}}}}, file)
            file.flush()
            file.seek(0)
            with self.mock_tagging_by_query_endpoint() as resp:
                runner = CliRunner()
                runner.invoke(cli, ['tagging-by-query', '--elasticsearch-url', self.elasticsearch_url,
                                    '--datashare-project', self.datashare_project, file.name])
                self.assertIsNone(resp.calls[0].request.headers.get('Authorization'))

    def test_apikey_header_is_sent_while_tagging_by_query_with_cli(self):
        with NamedTemporaryFile(mode='w+') as file:
            json.dump({"tag-name": {"query": {"match_all": {}}}}, file)
            file.flush()
            file.seek(0)
            with self.mock_tagging_by_query_endpoint() as resp:
                runner = CliRunner()
                runner.invoke(cli, ['tagging-by-query', '--elasticsearch-url', self.elasticsearch_url,
                                    '--datashare-project', self.datashare_project, '--apikey', 'my_api_key',
                                    file.name])
                self.assertEqual(resp.calls[0].request.headers['Authorization'], 'bearer my_api_key')

    def test_apikey_header_is_NOT_sent_while_tag_cleaning_by_query_with_cli(self):
        with NamedTemporaryFile() as file:
            file.write(b'{"query": {"match_all": {}}}')
            file.flush()
            file.seek(0)
            with self.mock_tagging_by_query_endpoint() as resp:
                runner = CliRunner()
                runner.invoke(cli, ['clean-tags-by-query', '--elasticsearch-url', self.elasticsearch_url,
                                    '--datashare-project', self.datashare_project, '--query',
                                    '@' + file.name])
                self.assertIsNone(resp.calls[0].request.headers.get('Authorization'))

    def test_apikey_header_is_sent_while_tag_cleaning_by_query_with_cli(self):
        self.datashare_client.index(index=self.datashare_project, document={'content': 'content', 'tags': ['tag']},
                                    id='id')
        with self.mock_tagging_by_query_endpoint() as resp:
            runner = CliRunner()
            runner.invoke(cli, ['clean-tags-by-query', '--elasticsearch-url', self.elasticsearch_url,
                                '--datashare-project', self.datashare_project, '--apikey', 'my_api_key', '--query',
                                '{"query": {"ids": {"values": ["id"]}}}'])
            self.assertEqual(resp.calls[0].request.headers['Authorization'], 'bearer my_api_key')

    def test_apikey_header_is_NOT_sent_while_downloading_with_cli(self):
        self.datashare_client.index(index=self.datashare_project, document={'type': 'Document', 'content': 'content',
                                                                            'tags': ['tag']}, id='id')
        with self.mock_download_endpoint() as resp:
            runner = CliRunner()
            runner.invoke(cli, ['download', '--elasticsearch-url', self.elasticsearch_url, '--datashare-url',
                                self.datashare_url, '--datashare-project',
                                self.datashare_project, '--query', '*'])
            self.assertIsNone(resp.calls[0].request.headers.get('Authorization'))

    def test_apikey_header_is_sent_while_downloading_with_cli(self):
        self.datashare_client.index(index=self.datashare_project, document={'type': 'Document', 'content': 'content',
                                                                            'tags': ['tag']}, id='id')
        with self.mock_download_endpoint() as resp:
            runner = CliRunner()
            runner.invoke(cli, ['download', '--elasticsearch-url', self.elasticsearch_url, '--datashare-url',
                                self.datashare_url, '--datashare-project',
                                self.datashare_project, '--apikey', 'my_api_key', '--query', '*'])
            self.assertEqual(resp.calls[0].request.headers['Authorization'], 'bearer my_api_key')



import csv
import json
import operator
import sys

from contextlib import contextmanager
from requests.exceptions import HTTPError
from urllib3.exceptions import ProtocolError
from tqdm import tqdm
from time import sleep

from tarentula.datashare_client import DatashareClient
from tarentula.logger import logger


class ExportByQuery:
    def __init__(self,
                 datashare_url: str = 'http://localhost:8080',
                 datashare_project: str = 'local-datashare',
                 output_file: str = 'tarentula_documents.csv',
                 query: str = '*',
                 throttle: int = 0,
                 cookies: str = '',
                 apikey: str = None,
                 elasticsearch_url: str = None,
                 scroll: str = None,
                 source: str = 'contentType,contentLength:0,extractionDate,path',
                 once: bool = False,
                 traceback: bool = False,
                 progressbar: bool = True,
                 type: str = 'Document'):
        self.datashare_url = datashare_url
        self.datashare_project = datashare_project
        self.query = query
        self.output_file = output_file
        self.throttle = throttle
        self.cookies_string = cookies
        self.apikey = apikey
        self.once = once
        self.traceback = traceback
        self.progressbar = progressbar
        self.scroll = scroll
        self.source = source
        self.type = type
        try:
            self.datashare_client = DatashareClient(datashare_url,
                                                    elasticsearch_url,
                                                    datashare_project,
                                                    cookies,
                                                    apikey)
        except (ConnectionRefusedError, ConnectionError):
            logger.critical('Unable to connect to Datashare', exc_info=self.traceback)
            exit()

    @property
    def query_body(self):
        if self.query.startswith('@'):
            return self.query_body_from_file
        else:
            return self.query_body_from_string

    @property
    def query_body_from_string(self):
        return {
            "query": {
                "bool": {
                    "must": [
                        {
                            "match": {
                                "type": self.type
                            }
                        },
                        {
                            "query_string": {
                                "query": self.query
                            }
                        }
                    ]
                }
            }
        }

    @property
    def query_body_from_file(self):
        with open(self.query[1:]) as json_file:
            query_body = json.load(json_file)
        return query_body

    @property
    def no_progressbar(self):
        return not self.progressbar

    @property
    def source_fields(self):
        return [ self.source_field_params(f) for f in self.source.split(',') ]

    @property
    def source_fields_names(self):
        return [ field.pop(0) for field in self.source_fields ]

    def source_field_params(self, field):
        field_params = field.strip().split(':')
        field_name = field_params[0]
        field_default = field_params[1] if len(field_params) > 1 else ''
        return [field_name, field_default]

    def sleep(self):
        sleep(self.throttle / 1000)

    def count_matches(self):
        index = self.datashare_project
        return self.datashare_client.count(index=index, query=self.query_body).get('count')

    def log_matches(self):
        index = self.datashare_project
        count = self.count_matches()
        logger.info('%s matching document(s) in %s' % (count, index))
        return count

    def scan_or_query_all(self):
        index = self.datashare_project
        source = ["path"] + self.source_fields_names
        if self.scroll is None:
            logger.info('Searching document(s) metadata in %s' % index)
            return self.datashare_client.query_all(index=index, query=self.query_body, source=source)
        else:
            logger.info('Scrolling over document(s) metadata in %s' % index)
            return self.datashare_client.scan_all(index=index, query=self.query_body, source=source, scroll=self.scroll)

    def document_default_values(self, document, number):
        index = self.datashare_project
        id = document.get('_id')
        routing = document.get('_routing', id)
        url = self.datashare_client.document_url(index, id, routing)
        return { 'query': self.query, 'documentUrl': url, 'documentId': id,
                 'rootId': routing, 'documentNumber': number }

    def document_source_values(self, document):
        source_values = {}
        source = document.get('_source', {})
        for [name, default] in self.source_fields:
            # Get the nested value for `name` (it can be a path, ie: metadata.tika_metadata_author)
            source_values[name] = source
            for key in name.split('.'):
                try:
                    source_values[name] = source_values[name][key]
                except (KeyError, TypeError):
                    source_values[name] = default
        return source_values

    def save_indexed_document(self, csvwriter, document, document_number):
        default_values = self.document_default_values(document, document_number)
        source_values = self.document_source_values(document)
        csvwriter.writerow({ **default_values, **source_values })

    @contextmanager
    def create_csv_file(self):
        with open(self.output_file, 'w', newline='') as csv_file:
            fieldnames = ['query', 'documentUrl', 'documentId', 'rootId', 'documentNumber']
            fieldnames += self.source_fields_names
            writer = csv.DictWriter(csv_file, fieldnames=fieldnames)
            writer.writeheader()
            yield writer

    def start(self):
        count = self.log_matches()
        try:
            documents = self.scan_or_query_all()
            pbar = tqdm(documents, total=count, desc='Exporting %s document(s)' % count,
                        file=sys.stdout, disable=self.no_progressbar)
            with self.create_csv_file() as csvwriter:
                for index, document in enumerate(pbar):
                    try:
                        self.save_indexed_document(csvwriter, document, index)
                        logger.info('Saved document %s' % document.get('_id', None))
                        self.sleep()
                    except HTTPError:
                        logger.error('Unable to export document %s' % document.get('_id', None),
                                        exc_info=self.traceback)
                logger.info('Written documents metadata in %s' % self.output_file)
        except ProtocolError:
            logger.error('Exception while exporting documents', exc_info=self.traceback)



import logging

from click.testing import CliRunner
from tempfile import NamedTemporaryFile

from tarentula.cli import cli
from tarentula.logger import add_stdout_handler
from tarentula.tag_cleaning_by_query import TagsCleanerByQuery
from tests.test_abstract import TestAbstract


class TestTagsCleanerByQuery(TestAbstract):
    def test_cli_is_wired_on_tags_cleaner(self):
        add_stdout_handler(level=logging.INFO)
        runner = CliRunner()
        result = runner.invoke(cli, ['clean-tags-by-query', '--datashare-project', self.datashare_project,
                                     '--elasticsearch-url', self.elasticsearch_url])
        self.assertIn('This action will remove all tags for documents matching query', result.output)

    def test_cli_is_wired_on_tags_cleaner_with_no_query_option_selects_all(self):
        self.datashare_client.index(index=self.datashare_project, document={'content': 'content', 'tags': ['tag']},
                                    id='id')
        add_stdout_handler(level=logging.INFO)
        runner = CliRunner()
        result = runner.invoke(cli, ['clean-tags-by-query', '--datashare-project', self.datashare_project,
                                     '--elasticsearch-url', self.elasticsearch_url])
        self.assertIn('updated 1 documents', result.output)

    def test_cli_is_wired_on_tags_cleaner_with_query_option(self):
        self.datashare_client.index(index=self.datashare_project, document={'content': 'content', 'tags': ['tag']},
                                    id='id')
        add_stdout_handler(level=logging.INFO)
        runner = CliRunner()
        result = runner.invoke(cli, ['clean-tags-by-query', '--datashare-project', self.datashare_project,
                                     '--elasticsearch-url', self.elasticsearch_url,
                                     '--query', '{"query": {"ids": {"values": ["id"]}}}'])
        self.assertIn('updated 1 documents', result.output)

    def test_cli_is_wired_on_tags_cleaner_with_background_task(self):
        add_stdout_handler(level=logging.INFO)
        runner = CliRunner()
        result = runner.invoke(cli, ['clean-tags-by-query', '--datashare-project', self.datashare_project,
                                     '--elasticsearch-url', self.elasticsearch_url, '--no-wait-for-completion'])
        self.assertIn('task created:', result.output)

    def test_cli_is_wired_on_tags_cleaner_with_query_option_in_a_file(self):
        self.datashare_client.index(index=self.datashare_project, document={'content': 'content', 'tags': ['tag']},
                                    id='id')
        add_stdout_handler(level=logging.INFO)
        with NamedTemporaryFile() as file:
            file.write(b'{"query": {"match_all": {}}}')
            file.flush()
            file.seek(0)
            runner = CliRunner()
            result = runner.invoke(cli, ['clean-tags-by-query', '--datashare-project', self.datashare_project,
                                         '--elasticsearch-url', self.elasticsearch_url, '--query',
                                         '@' + file.name])
            self.assertIn('updated 1 documents', result.output)

    def test_clean_tags(self):
        self.datashare_client.index(index=self.datashare_project, document={'content': 'content',
                                                                            'tags': ['tag1', 'tag2']}, id='id')
        TagsCleanerByQuery(self.datashare_project, self.elasticsearch_url).start()
        document = self.datashare_client.document(self.datashare_project, id='id')['_source']
        self.assertListEqual(document['tags'], [])



import json
import sys
import requests

from http.cookies import SimpleCookie
from requests.exceptions import HTTPError, ConnectionError
from time import sleep
from tqdm import tqdm

from tarentula.logger import logger


class TaggerByQuery:
    def __init__(self,
                 datashare_project: str = '',
                 elasticsearch_url: str = '',
                 json_path: str = '',
                 throttle: int = 0,
                 cookies: str = '',
                 apikey: str = None,
                 progressbar: bool = True,
                 traceback: bool = False,
                 wait_for_completion: bool = True,
                 scroll_size: int = 1000):
        self.datashare_project = datashare_project
        self.elasticsearch_url = elasticsearch_url
        self.cookies_string = cookies
        self.apikey = apikey
        self.throttle = throttle
        self.json_path = json_path
        self.traceback = traceback
        self.progressbar = progressbar
        self.wait_for_completion = wait_for_completion
        self.scroll_size = scroll_size

    @property
    def no_progressbar(self):
        return not self.progressbar

    @property
    def cookies(self):
        cookies = SimpleCookie()
        try:
            cookies.load(self.cookies_string)
            return {key: morsel.value for (key, morsel) in cookies.items()}
        except (TypeError, AttributeError):
            return {}

    @property
    def tags(self):
        json_file = open(self.json_path, 'r')
        tags = json.loads(json_file.read())
        json_file.close()
        return tags

    @property
    def tagging_by_query_endpoint(self):
        url_template = '{elasticsearch_url}/{datashare_project}/_update_by_query?conflicts=proceed'
        return url_template.format(elasticsearch_url=self.elasticsearch_url, datashare_project=self.datashare_project)

    def sleep(self):
        sleep(self.throttle / 1000)

    def task_url(self, task):
        url_template = '{elasticsearch_url}/_tasks/{task}'
        return url_template.format(elasticsearch_url=self.elasticsearch_url, task=task)

    def tag_documents(self, tag, query):
        query = {
            "script": {
                "source": """
                    if( !ctx._source.containsKey("tags") ) {
                        ctx._source.tags = [];
                    }
                    if( !ctx._source.tags.contains(params.tag) ) {
                        ctx._source.tags.add(params.tag);
                    }
                """,
                "lang": "painless",
                "params": {
                    "tag": tag
                },
            },
            **query
        }
        params = {
            "wait_for_completion": str(self.wait_for_completion).lower(),
            "scroll_size": self.scroll_size,
        }
        result = requests.post(self.tagging_by_query_endpoint, params=params, json=query, cookies=self.cookies,
                               headers=None if self.apikey is None else {'Authorization': 'bearer %s' % self.apikey})
        result.raise_for_status()
        return result

    @property
    def tags_count(self):
        return len(self.tags.keys())

    def start(self):
        count = self.tags_count
        pbar = tqdm(self.tags.items(), total=count, desc="This action will add %s tag(s)" % count, file=sys.stderr,
                    disable=self.no_progressbar)
        for (tag, query) in pbar:
            try:
                tqdm.write('Adding "%s" tag' % tag)
                result = self.tag_documents(tag, query).json()
                if self.wait_for_completion:
                    tqdm.write('└── documents updated in %sms' % result['took'])
                    logger.info('Documents tagged with [%s] in %sms' % (tag, result['took']))
                else:
                    tqdm.write('└── task created: %s' % self.task_url(result['task']))
                    logger.info('Task [%s] created for tag [%s]' % (result['task'], tag))
                self.sleep()
            except (HTTPError, ConnectionError):
                logger.error('Unable to add tag [%s] (connection error)' % tag, exc_info=self.traceback)



import sys

from setuptools import setup, find_packages
from tarentula import __version__

py_version = sys.version_info[:2]
if py_version < (3, 6):
    raise Exception("tarentula requires Python >= 3.6.")

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(name='tarentula',
      version=__version__,
      packages=find_packages(),
      description="Cli toolbelt for Datashare.",
      use_pipfile=True,
      long_description=long_description,
      long_description_content_type="text/markdown",
      url="https://github.com/icij/datashare-tarentula",
      test_suite='nose.collector',
      tests_require=['nose', 'responses'],
      include_package_data=True,
      keywords=['datashare', 'api', 'text-mining', 'elasticsearch'],
      classifiers=[
          "Programming Language :: Python :: 3",
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: 3.7",
          "Programming Language :: Python :: 3.8",
          "Intended Audience :: Developers",
          "License :: OSI Approved :: GNU Affero General Public License v3",
          "Operating System :: OS Independent",
          "Topic :: Scientific/Engineering :: Information Analysis",
      ],
      python_requires='>=3.6',
      entry_points='''
        [console_scripts]
        tarentula=tarentula.cli:cli
        graph_es=tarentula.graph_realtime:graph
    ''')



from click.testing import CliRunner

from .test_abstract import TestAbstract
from tarentula.cli import cli
from tarentula import __version__

class TestVersion(TestAbstract):

    def test_version(self):
        runner = CliRunner()
        result = runner.invoke(cli, ['--version'])
        self.assertIn('v%s' % __version__, result.output)

    def test_version_help(self):
        runner = CliRunner()
        result = runner.invoke(cli, ['--help'])
        self.assertIn('Show the version and exit.', result.output)



import os
from unittest import TestCase

from tarentula.datashare_client import DatashareClient
from tarentula.graph_realtime import GraphRealTime


class TestGraphRealTime(TestCase):
    es_url = os.environ.get('TEST_ELASTICSEARCH_URL', 'http://elasticsearch:9200')
    ds_client = None

    @classmethod
    def setUpClass(cls) -> None:
        cls.ds_client = DatashareClient(elasticsearch_url=cls.es_url)
        cls.ds_client.create('test-datashare')

    @classmethod
    def tearDownClass(cls) -> None:
        cls.ds_client.delete_index('test-datashare')

    def test_field(self):
        xs = []
        ys = []
        GraphRealTime(query='{"query":{"match_all":{}}}', elasticsearch_url=self.es_url,
                      index='test-datashare', field='hits.total.value', refresh_interval=5, xs_param=xs, ys_param=ys).add_point(0)

        self.assertEqual(ys, [0])



import requests

from contextlib import contextmanager
from datetime import datetime
from http.cookies import SimpleCookie
from uuid import uuid4


def urljoin(*args):
    return '/'.join(s.strip('/') for s in args if s is not None)


DATASHARE_DEFAULT_PROJECT = 'local-datashare'


class DatashareClient:
    def __init__(self, datashare_url='http://localhost:8080', elasticsearch_url=None,
                 datashare_project=DATASHARE_DEFAULT_PROJECT, cookies='', apikey=None):
        self.datashare_url = datashare_url
        self.cookies_string = cookies
        self.apikey = apikey
        self.elasticsearch_url = elasticsearch_url
        # Create the datashare default index
        self.create(datashare_project)

    @property
    def cookies(self):
        cookies = SimpleCookie()
        try:
            cookies.load(self.cookies_string)
            return {key: morsel.value for (key, morsel) in cookies.items()}
        except (TypeError, AttributeError):
            return {}

    @property
    def headers(self):
        if self.apikey is None:
            return None
        else:
            return {'Authorization': 'bearer %s' % self.apikey}

    @property
    def elasticsearch_host(self):
        if self.elasticsearch_url is not None:
            return self.elasticsearch_url
        else:
            # @see https://github.com/ICIJ/datashare/wiki/Datashare-API
            return urljoin(self.datashare_url, '/api/index/search/')

    def create(self, index=DATASHARE_DEFAULT_PROJECT):
        url = urljoin(self.datashare_url, '/api/index/', index)
        return requests.put(url)

    def index(self, index=DATASHARE_DEFAULT_PROJECT, document=None, id=None, routing=None):
        if document is None:
            document = {}
        params = {'routing': routing}
        # Clone the document to perform changes
        document = dict(document)
        # Elasticsearch doesn't allow passing the _id as a property in the document
        if '_id' in document:
            document.pop('_id', None)
        if '_routing' in document:
            document.pop('_routing', None)
        # When no id is provided, we use POST method (to create the resource)
        if 'content' in document:
            content_length = len(document.get('content', ''))
            document.update({'contentLength': content_length})
        now = datetime.now()
        extraction_date = now.strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3] + 'Z'
        document.update({'extractionDate': extraction_date})
        if id is None:
            url = urljoin(self.elasticsearch_url, index, '/_doc?refresh')
            result = requests.post(url, json=document, params=params)
        # When an id is provided, we use PUT method (to update the resource)
        else:
            if routing is None:
                query_params = '?refresh'
            else:
                query_params = '?refresh&routing=' + routing
            url = urljoin(self.elasticsearch_url, index, '/_doc/', id, query_params)
            result = requests.put(url, json=document, params=params)
        result.raise_for_status()
        return result.json().get('_id')

    def delete(self, index=DATASHARE_DEFAULT_PROJECT, id=None):
        url = urljoin(self.elasticsearch_url, index, '/_doc/', id, '?refresh')
        return requests.delete(url)

    def refresh(self, index=DATASHARE_DEFAULT_PROJECT):
        url = urljoin(self.elasticsearch_url, index, '/_refresh')
        return requests.post(url)

    def delete_index(self, index):
        url = urljoin(self.elasticsearch_url, index)
        return requests.delete(url)

    def delete_all(self, index):
        url = urljoin(self.elasticsearch_url, index, '_delete_by_query')
        body = {"query": {"match_all": {}}}
        params = {"conflicts": "proceed", "refresh": True}
        return requests.post(url, json=body, params=params)

    def reindex(self, source=DATASHARE_DEFAULT_PROJECT, dest=None, size=1):
        # Create a default destination index name
        if dest is None:
            dest = '%s-copy-%s' % (source, uuid4().hex[:6])
        # Source index must at least have one document
        document_id = self.index(source, document={"content": "This is a temporary document", "tags": ["tmp"]})
        # Copy everything
        json = {"source": {"index": source}, "dest": {"index": dest}, "size": size}
        # Send the request to elasticsearch
        url = urljoin(self.elasticsearch_url, '_reindex')
        result = requests.post(url + '?refresh', json=json)
        # Delete the dummy docs
        self.delete(source, document_id)
        self.delete(dest, document_id)
        self.delete_all(dest)
        # Return the dest name
        return dest if result.status_code == requests.codes.ok else None

    def query(self, index=DATASHARE_DEFAULT_PROJECT, query={}, q=None, source=None, scroll=None, **kwargs):
        local_query = {"sort": {"_id": "asc"}, **query, **kwargs}
        if source is not None:
            local_query.update({'_source': source})
        url = urljoin(self.elasticsearch_host, index, '/_doc/_search')
        response = requests.post(url, params={"q": q, "scroll": scroll},
                                 json=local_query,
                                 headers=self.headers,
                                 cookies=self.cookies)
        response.raise_for_status()
        return response.json()

    def scroll(self, scroll_id, scroll=None):
        url = urljoin(self.elasticsearch_host, '/_search/scroll')
        body = {"scroll_id": scroll_id, "scroll": scroll}
        response = requests.post(url, json=body,
                                 cookies=self.cookies,
                                 headers=self.headers)
        response.raise_for_status()
        return response.json()

    def scan_all(self, scroll='10m', **kwargs):
        response = self.query(scroll=scroll, **kwargs)
        while len(response['hits']['hits']) > 0:
            for item in response['hits']['hits']:
                yield item
            if '_scroll_id' not in response:
                break
            scroll_id = response['_scroll_id']
            response = self.scroll(scroll_id, scroll)

    def query_all(self, **kwargs):
        response = self.query(**kwargs)
        while len(response['hits']['hits']) > 0:
            for item in response['hits']['hits']:
                yield item
            search_after = response['hits']['hits'][-1]['sort']
            response = self.query(search_after=search_after, **kwargs)

    def count(self, index=DATASHARE_DEFAULT_PROJECT, query={}):
        url = urljoin(self.elasticsearch_host, index, '_count')
        return requests.post(url, json=query,
                             cookies=self.cookies,
                             headers=self.headers).json()

    def document(self, index=DATASHARE_DEFAULT_PROJECT, id=None, routing=None, source=None):
        url = urljoin(self.elasticsearch_host, index, '/_doc/', id)
        params = {'routing': routing, '_source': source}
        return requests.get(url, params=params,
                            cookies=self.cookies,
                            headers=self.headers).json()

    def download(self, index=DATASHARE_DEFAULT_PROJECT, id=None, routing=None):
        routing = routing or id
        url = urljoin(self.datashare_url, 'api', index, '/documents/src', id)
        return requests.get(url, params={'routing': routing},
                            cookies=self.cookies,
                            headers=self.headers,
                            stream=True)

    def document_url(self, index=DATASHARE_DEFAULT_PROJECT, id='', routing=None):
        routing = id if routing is None else routing
        return urljoin(self.datashare_url, f'#/d/{index}/{id}/{routing}')

    @contextmanager
    def temporary_project(self, source=DATASHARE_DEFAULT_PROJECT, delete=True):
        project = None
        try:
            project = self.reindex(source)
            yield project
        finally:
            if delete and project is not None:
                self.delete_index(project)
        return project
